# Copyright (c) 2012 - 2014 EMC Corporation, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
VNX CLI on iSCSI.
"""

import os
import time

from oslo.config import cfg

from cinder import exception
from cinder.openstack.common import log as logging
from cinder.openstack.common import loopingcall
from cinder.openstack.common import processutils
from cinder import utils
from cinder.volume.drivers.san import san
from cinder.volume import volume_types

LOG = logging.getLogger(__name__)

CONF = cfg.CONF
VERSION = '02.00.00'

loc_opts = [
    cfg.StrOpt('naviseccli_path',
               default='',
               help='Naviseccli Path'),
    cfg.StrOpt('storage_vnx_pool_name',
               default=None,
               help='ISCSI pool name'),
    cfg.IntOpt('default_timeout',
               default=20,
               help='Default Time Out For CLI operations in minutes'),
    cfg.IntOpt('max_luns_per_storage_group',
               default=256,
               help='Default max number of LUNs in a storage group'), ]

CONF.register_opts(loc_opts)


class EMCVnxCli(object):
    """This class defines the functions to use the native CLI functionality."""

    stats = {'driver_version': VERSION,
             'free_capacity_gb': 'unknown',
             'reserved_percentage': 0,
             'storage_protocol': None,
             'total_capacity_gb': 'unknown',
             'vendor_name': 'EMC',
             'volume_backend_name': None}

    def __init__(self, prtcl, configuration=None):

        self.protocol = prtcl
        self.configuration = configuration
        self.configuration.append_config_values(loc_opts)
        self.configuration.append_config_values(san.san_opts)
        self.storage_ip = self.configuration.san_ip
        self.storage_username = self.configuration.san_login
        self.storage_password = self.configuration.san_password

        self.pool_name = self.configuration.storage_vnx_pool_name
        if not self.pool_name:
            msg = (_('Pool name is not specified.'))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        self.timeout = self.configuration.default_timeout
        self.max_luns = self.configuration.max_luns_per_storage_group
        self.hlu_set = set(xrange(1, self.max_luns + 1))
        self.navisecclipath = self.configuration.naviseccli_path
        self.cli_prefix = (self.navisecclipath, '-address', self.storage_ip)
        self.cli_credentials = ()
        self.wait_interval = 3

        # if there is a username/password provided, use those in the cmd line
        if self.storage_username is not None and \
                self.storage_password is not None:
            self.cli_credentials += ('-user', self.storage_username,
                                     '-password', self.storage_password,
                                     '-scope', '0')

        # Checking for existence of naviseccli tool
        if not os.path.exists(self.navisecclipath):
            msg = (_('Could not find NAVISECCLI tool.'))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Testing the naviseccli setup
        query_list = ("storagepool", "-list",
                      "-name", self.pool_name, "-state")
        out, rc = self._cli_execute(*query_list)
        if rc != 0:
            LOG.error(_("Failed to find pool %s"), self.pool_name)
            raise exception.VolumeBackendAPIException(data=out)

    def _cli_execute(self, *cmd, **kwargv):
        if "check_exit_code" not in kwargv:
            kwargv["check_exit_code"] = True
        rc = 0
        try:
            out, _err = utils.execute(*(self.cli_prefix +
                                      self.cli_credentials + cmd), **kwargv)
        except processutils.ProcessExecutionError as pe:
            rc = pe.exit_code
            out = pe.stdout + pe.stderr
        return out, rc

    def create_volume(self, volume):
        """Creates a EMC volume."""

        LOG.debug(_('Entering create_volume.'))
        volumesize = volume['size']
        volumename = volume['name']

        LOG.info(_('Create Volume: %(volume)s  Size: %(size)s')
                 % {'volume': volumename,
                    'size': volumesize})

        # defining CLI command
        thinness = self._get_provisioning_by_volume(volume)

        # executing CLI command to create volume
        LOG.debug(_('Create Volume: %(volumename)s')
                  % {'volumename': volumename})

        lun_create = ('lun', '-create',
                      '-type', thinness,
                      '-capacity', volumesize,
                      '-sq', 'gb',
                      '-poolName', self.pool_name,
                      '-name', volumename)
        out, rc = self._cli_execute(*lun_create)
        LOG.debug(_('Create Volume: %(volumename)s  Return code: %(rc)s')
                  % {'volumename': volumename,
                     'rc': rc})
        if rc == 4:
            LOG.warn(_('Volume %s already exists'), volumename)
        elif rc != 0:
            msg = (_('Failed to create %(volumename)s: %(out)s') %
                   {'volumename': volumename, 'out': out})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # wait for up to a minute to verify that the LUN has progressed
        # to Ready state
        def _wait_for_lun_ready(volumename, start_time):
            # executing cli command to check volume
            command_to_verify = ('lun', '-list', '-name', volumename)
            out, rc = self._cli_execute(*command_to_verify)
            if rc == 0 and out.find("Ready") > -1:
                raise loopingcall.LoopingCallDone()
            if int(time.time()) - start_time > self.timeout * 60:
                msg = (_('LUN %s failed to become Ready'), volumename)
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

        timer = loopingcall.FixedIntervalLoopingCall(
            _wait_for_lun_ready, volumename, int(time.time()))
        timer.start(interval=self.wait_interval).wait()

    def delete_volume(self, volume):
        """Deletes an EMC volume."""

        LOG.debug(_('Entering delete_volume.'))
        volumename = volume['name']
        # defining CLI command
        lun_destroy = ('lun', '-destroy',
                       '-name', volumename,
                       '-forceDetach', '-o')

        # executing CLI command to delete volume
        out, rc = self._cli_execute(*lun_destroy)
        LOG.debug(_('Delete Volume: %(volumename)s  Output: %(out)s')
                  % {'volumename': volumename, 'out': out})
        if rc not in (0, 9):
            msg = (_('Failed to destroy %s'), volumename)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

    def extend_volume(self, volume, new_size):
        """Extends an EMC volume."""

        LOG.debug(_('Entering extend_volume.'))
        volumename = volume['name']

        # defining CLI command
        lun_expand = ('lun', '-expand',
                      '-name', volumename,
                      '-capacity', new_size,
                      '-sq', 'gb',
                      '-o', '-ignoreThresholds')

        # executing CLI command to extend volume
        out, rc = self._cli_execute(*lun_expand)

        LOG.debug(_('Extend Volume: %(volumename)s  Output: %(out)s')
                  % {'volumename': volumename,
                     'out': out})
        if rc == 97:
            msg = (_('The LUN cannot be expanded or shrunk because '
                   'it has snapshots. Command to extend the specified '
                   'volume failed.'))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        if rc != 0:
            msg = (_('Failed to expand %s'), volumename)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

    def update_volume_status(self):
        """Retrieve status info."""
        LOG.debug(_("Updating volume status"))

        poolname = self.pool_name
        pool_list = ('storagepool', '-list',
                     '-name', poolname,
                     '-userCap', '-availableCap')
        out, rc = self._cli_execute(*pool_list)
        if rc == 0:
            pool_details = out.split('\n')
            self.stats['total_capacity_gb'] = float(
                pool_details[3].split(':')[1].strip())
            self.stats['free_capacity_gb'] = float(
                pool_details[5].split(':')[1].strip())
        else:
            msg = (_('Failed to list %s'), poolname)
            LOG.error(msg)

        return self.stats

    def create_export(self, context, volume):
        """Driver entry point to get the export info for a new volume."""
        volumename = volume['name']

        device_id = self._find_lun_id(volumename)

        LOG.debug(_('create_export: Volume: %(volume)s  Device ID: '
                  '%(device_id)s')
                  % {'volume': volumename,
                     'device_id': device_id})

        return {'provider_location': device_id}

    def _find_lun_id(self, volumename):
        """Returns the LUN of a volume."""

        lun_list = ('lun', '-list', '-name', volumename)

        out, rc = self._cli_execute(*lun_list)
        if rc == 0:
            vol_details = out.split('\n')
            lun = vol_details[0].split(' ')[3]
        else:
            msg = (_('Failed to list %s'), volumename)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        return lun

    def create_snapshot(self, snapshot):
        """Creates a snapshot."""
        LOG.debug(_('Entering create_snapshot.'))
        snapshotname = snapshot['name']
        volumename = snapshot['volume_name']
        LOG.info(_('Create snapshot: %(snapshot)s: volume: %(volume)s')
                 % {'snapshot': snapshotname,
                    'volume': volumename})

        volume_lun = self._find_lun_id(volumename)

        # defining CLI command
        snap_create = ('snap', '-create',
                       '-res', volume_lun,
                       '-name', snapshotname,
                       '-allowReadWrite', 'yes')
        # executing CLI command to create snapshot
        out, rc = self._cli_execute(*snap_create)

        LOG.debug(_('Create Snapshot: %(snapshotname)s  Unity: %(out)s')
                  % {'snapshotname': snapshotname,
                     'out': out})
        if rc != 0:
            msg = (_('Failed to create snap %s'), snapshotname)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

    def delete_snapshot(self, snapshot):
        """Deletes a snapshot."""
        LOG.debug(_('Entering delete_snapshot.'))

        snapshotname = snapshot['name']
        volumename = snapshot['volume_name']
        LOG.info(_('Delete Snapshot: %(snapshot)s: volume: %(volume)s')
                 % {'snapshot': snapshotname,
                    'volume': volumename})

        def _wait_for_snap_delete(snapshot, start_time):
            # defining CLI command
            snapshotname = snapshot['name']
            volumename = snapshot['volume_name']
            snap_destroy = ('snap', '-destroy', '-id', snapshotname, '-o')
            # executing CLI command
            out, rc = self._cli_execute(*snap_destroy)

            LOG.debug(_('Delete Snapshot: Volume: %(volumename)s  Snapshot: '
                      '%(snapshotname)s  Output: %(out)s')
                      % {'volumename': volumename,
                         'snapshotname': snapshotname,
                         'out': out})

            if rc not in [0, 9, 5]:
                if rc == 13:
                    if int(time.time()) - start_time < \
                            self.timeout * 60:
                        LOG.info(_('Snapshot %s is in use'), snapshotname)
                    else:
                        msg = (_('Failed to destroy %s '
                               ' because snapshot is in use.'), snapshotname)
                        LOG.error(msg)
                        raise exception.SnapshotIsBusy(data=msg)
                else:
                    msg = (_('Failed to destroy %s'), snapshotname)
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)
            else:
                raise loopingcall.LoopingCallDone()

        timer = loopingcall.FixedIntervalLoopingCall(
            _wait_for_snap_delete, snapshot, int(time.time()))
        timer.start(interval=self.wait_interval).wait()

    def create_volume_from_snapshot(self, volume, snapshot):
        """Creates a volume from a snapshot."""

        LOG.debug(_('Entering create_volume_from_snapshot.'))

        snapshotname = snapshot['name']
        source_volume_name = snapshot['volume_name']
        volumename = volume['name']
        volumesize = snapshot['volume_size']

        destvolumename = volumename + 'dest'

        # Create a mount point, migrate data from source (snapshot) to
        # destination volume.  The destination volume is the only new volume
        # to be created here.
        LOG.info(_('Creating Destination Volume : %s ') % (destvolumename))

        poolname = self.pool_name
        thinness = self._get_provisioning_by_volume(volume)
        # defining CLI command
        lun_create = ('lun', '-create', '-type', thinness,
                      '-capacity', volumesize, '-sq', 'gb',
                      '-poolName', poolname,
                      '-name', destvolumename)
        # executing CLI command
        out, rc = self._cli_execute(*lun_create)

        LOG.debug(_('Create temporary Volume: %(volumename)s  '
                  'Output : %(out)s')
                  % {'volumename': destvolumename, 'out': out})

        if rc != 0:
            msg = (_('Command to create the destination volume failed'))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # defining CLI command
        smp_create = ('lun', '-create', '-type', 'Snap',
                      '-primaryLunName', source_volume_name,
                      '-name', volumename)

        # executing CLI command
        out, rc = self._cli_execute(*smp_create)
        LOG.debug(_('Create mount point : Volume: %(volumename)s  '
                  'Source Volume: %(sourcevolumename)s  Output: %(out)s')
                  % {'volumename': volumename,
                     'sourcevolumename': source_volume_name,
                     'out': out})

        if rc != 0:
            msg = (_('Failed to create SMP %s'), volumename)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # defining CLI command
        lun_attach = ('lun', '-attach',
                      '-name', volumename,
                      '-snapName', snapshotname)

        # executing CLI command
        out, rc = self._cli_execute(*lun_attach)
        LOG.debug(_('Attaching mount point Volume: %(volumename)s  '
                  'with  Snapshot: %(snapshotname)s  Output: %(out)s')
                  % {'volumename': volumename,
                     'snapshotname': snapshotname,
                     'out': out})

        if rc != 0:
            msg = (_('Failed to attach snapshotname %s'), snapshotname)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        source_vol_lun = self._find_lun_id(volumename)
        dest_vol_lun = self._find_lun_id(destvolumename)

        LOG.info(_('Migrating Mount Point Volume: %s ') % (volumename))

        # defining CLI command
        migrate_start = ('migrate', '-start',
                         '-source', source_vol_lun,
                         '-dest', dest_vol_lun,
                         '-rate', 'ASAP', '-o')

        # executing CLI command
        out, rc = self._cli_execute(*migrate_start)

        LOG.debug(_('Migrate Mount Point  Volume: %(volumename)s  '
                  'Output : %(out)s')
                  % {'volumename': volumename,
                     'out': out})

        if rc != 0:
            msg = (_('Failed to start migrating SMP %s'), volumename)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        def _wait_for_sync_status(volumename, start_time):
            lun_list = ('lun', '-list', '-name', volumename,
                        '-attachedSnapshot')
            out, rc = self._cli_execute(*lun_list)
            if rc == 0:
                vol_details = out.split('\n')
                snapshotname = vol_details[2].split(':')[1].strip()
            if (snapshotname == 'N/A'):
                raise loopingcall.LoopingCallDone()
            else:
                LOG.info(_('Waiting for the update on Sync status of %s'),
                         volumename)
                if int(time.time()) - start_time >= self.timeout * 60:
                    msg = (_('Failed to really migrate %s'), volumename)
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

        timer = loopingcall.FixedIntervalLoopingCall(
            _wait_for_sync_status, volumename, int(time.time()))
        timer.start(interval=self.wait_interval).wait()

    def create_cloned_volume(self, volume, src_vref):
        """Creates a clone of the specified volume."""

        source_volume_name = src_vref['name']
        volumesize = src_vref['size']
        snapshotname = source_volume_name + '-temp-snapshot'

        snapshot = {
            'name': snapshotname,
            'volume_name': source_volume_name,
            'volume_size': volumesize,
        }

        # Create temp Snapshot
        self.create_snapshot(snapshot)

        try:
            # Create volume
            self.create_volume_from_snapshot(volume, snapshot)
        except Exception:
            msg = (_('Failed to create cloned volume %s'), volume['name'])
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        finally:
            # Delete temp Snapshot
            self.delete_snapshot(snapshot)

    def get_storage_group(self, hostname):
        """Returns the storage group for the host node."""

        storage_groupname = hostname

        sg_list = ('storagegroup', '-list', '-gname', storage_groupname)

        out, rc = self._cli_execute(*sg_list)

        if rc != 0:
            LOG.debug(_('creating new storage group %s'), storage_groupname)

            sg_create = ('storagegroup', '-create',
                         '-gname', storage_groupname)
            out, rc = self._cli_execute(*sg_create)
            LOG.debug(_('Create new storage group : %(storage_groupname)s, '
                      'Output: %(out)s')
                      % {'storage_groupname': storage_groupname,
                         'out': out})

            if rc != 0:
                msg = (_('Failed to create SG %s'), storage_groupname)
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            # connecting the new storagegroup to the host
            connect_host = ('storagegroup', '-connecthost',
                            '-host', hostname,
                            '-gname', storage_groupname,
                            '-o')

            out, rc = self._cli_execute(*connect_host)
            LOG.debug(_('Connect storage group : %(storage_groupname)s ,'
                        'To Host : %(hostname)s, Output : %(out)s')
                      % {'storage_groupname': storage_groupname,
                         'hostname': hostname,
                         'out': out})

            if rc != 0:
                msg = (_('Failed to connect %s'), hostname)
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

        return hostname

    def find_device_details(self, volume, storage_group):
        """Returns the Host Device number for the volume."""

        allocated_lun_id = self._find_lun_id(volume["name"])
        host_lun_id = -1
        owner_sp = ""
        lun_map = {}

        sg_list = ('storagegroup', '-list', '-gname', storage_group)
        out, rc = self._cli_execute(*sg_list)
        if out.find('HLU/ALU Pairs') == -1:
            LOG.info(_('NO LUNs in the storagegroup : %s ')
                     % (storage_group))
        else:
            sg_details = out.split('HLU/ALU Pairs:')[1]
            sg_lun_details = sg_details.split('Shareable')[0]
            lun_details = sg_lun_details.split('\n')

            for data in lun_details:
                if data not in ['', '  HLU Number     ALU Number',
                                '  ----------     ----------']:
                    data = data.strip()
                    items = data.split(' ')
                    lun_map[int(items[len(items) - 1])] = int(items[0])
            for lun in lun_map.iterkeys():
                if lun == int(allocated_lun_id):
                    host_lun_id = lun_map[lun]
                    LOG.debug(_('Host Lun Id : %s') % (host_lun_id))
                    break

        # finding the owner SP for the LUN
        lun_list = ('lun', '-list', '-l', allocated_lun_id, '-owner')
        out, rc = self._cli_execute(*lun_list)
        if rc == 0:
            output = out.split('\n')
            owner_sp = output[2].split('Current Owner:  SP ')[1]
            LOG.debug(_('Owner SP : %s') % (owner_sp))

        device = {
            'hostlunid': host_lun_id,
            'ownersp': owner_sp,
            'lunmap': lun_map,
        }
        return device

    def _get_host_lun_id(self, host_lun_id_list):
        # Returns the host lun id for the LUN to be added
        # in the storage group.

        used_hlu_set = set(host_lun_id_list)
        for hlu in self.hlu_set - used_hlu_set:
            return hlu
        return None

    def _add_lun_to_storagegroup(self, volume, storage_group):

        storage_groupname = storage_group
        volumename = volume['name']
        allocated_lun_id = self._find_lun_id(volumename)
        count = 0
        while(count < 5):
            device_info = self.find_device_details(volume, storage_group)
            device_number = device_info['hostlunid']
            if device_number < 0:
                lun_map = device_info['lunmap']
                if lun_map:
                    host_lun_id_list = lun_map.values()

                    if len(host_lun_id_list) >= self.max_luns:
                        msg = (_('The storage group has reached the '
                               'maximum capacity of LUNs. '
                               'Command to add LUN for volume - %s '
                               'in storagegroup failed') % (volumename))
                        LOG.error(msg)
                        raise exception.VolumeBackendAPIException(data=msg)

                    host_lun_id = self._get_host_lun_id(host_lun_id_list)

                    if host_lun_id is None:
                        msg = (_('Unable to get new host lun id. Please '
                               'check if the storage group can accommodate '
                               'new LUN. '
                               'Command to add LUN for volume - %s '
                               'in storagegroup failed') % (volumename))
                        LOG.error(msg)
                        raise exception.VolumeBackendAPIException(data=msg)
                else:
                    host_lun_id = 1

                addhlu = ('storagegroup', '-addhlu', '-o',
                          '-gname', storage_groupname,
                          '-hlu', host_lun_id,
                          '-alu', allocated_lun_id)
                out, rc = self._cli_execute(*addhlu)
                LOG.debug(_('Add ALU %(alu)s to SG %(sg)s as %(hlu)s. '
                          'Output: %(out)s')
                          % {'alu': allocated_lun_id,
                             'sg': storage_groupname,
                             'hlu': host_lun_id,
                             'out': out})
                if rc == 0:
                    return host_lun_id
                if rc == 66:
                    LOG.warn(_('Requested Host LUN Number already in use'))
                count += 1
            else:
                LOG.warn(_('LUN was already added in the storage group'))
                return device_number

        if count == 5:
            msg = (_('Failed to add %s into SG') % (volumename))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

    def _remove_lun_from_storagegroup(self, device_number, storage_group):

        storage_groupname = storage_group
        removehlu = ('storagegroup', '-removehlu',
                     '-gname', storage_groupname,
                     '-hlu', device_number,
                     '-o')

        out, rc = self._cli_execute(*removehlu)

        LOG.debug(_('Remove %(hlu)s from SG %(sg)s. Output: %(out)s')
                  % {'hlu': device_number,
                     'sg': storage_groupname,
                     'out': out})
        if rc != 0:
            msg = (_('Failed to remove %(hlu)s from %(sg)s')
                   % {'hlu': device_number, 'sg': storage_groupname})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

    def initialize_connection(self, volume, connector):
        """Initializes the connection and returns connection info."""

        hostname = connector['host']
        storage_group = self.get_storage_group(hostname)

        device_number = self._add_lun_to_storagegroup(volume, storage_group)
        return device_number

    def terminate_connection(self, volume, connector):
        """Disallow connection from connector."""
        hostname = connector['host']
        storage_group = self.get_storage_group(hostname)
        device_info = self.find_device_details(volume, storage_group)
        device_number = device_info['hostlunid']
        if device_number < 0:
            LOG.error(_('Could not locate the attached volume.'))
        else:
            self._remove_lun_from_storagegroup(device_number, storage_group)

    def _find_iscsi_protocol_endpoints(self, device_sp):
        """Returns the iSCSI initiators for a SP."""

        initiator_address = []

        connection_getport = ('connection', '-getport', '-sp', device_sp)
        out, _rc = self._cli_execute(*connection_getport)
        output = out.split('SP:  ')

        for port in output:
            port_info = port.split('\n')
            if port_info[0] == device_sp:
                port_wwn = port_info[2].split('Port WWN:')[1].strip()
                initiator_address.append(port_wwn)

        LOG.debug(_('WWNs found for SP %(devicesp)s '
                  'are: %(initiator_address)s')
                  % {'devicesp': device_sp,
                     'initiator_address': initiator_address})

        return initiator_address

    def _get_volumetype_extraspecs(self, volume):
        specs = {}

        type_id = volume['volume_type_id']
        if type_id is not None:
            specs = volume_types.get_volume_type_extra_specs(type_id)

        return specs

    def _get_provisioning_by_volume(self, volume):
        # By default, the user can not create thin LUN without thin
        # provisioning enabler.
        thinness = 'NonThin'
        spec_id = 'storagetype:provisioning'

        specs = self._get_volumetype_extraspecs(volume)
        if specs and spec_id in specs:
            provisioning = specs[spec_id].lower()
            if 'thin' == provisioning:
                thinness = 'Thin'
            elif 'thick' != provisioning:
                LOG.warning(_('Invalid value of extra spec '
                            '\'storagetype:provisioning\': %(provisioning)s')
                            % {'provisioning': specs[spec_id]})
        else:
            LOG.info(_('No extra spec \'storagetype:provisioning\' exist'))

        return thinness



# Copyright (c) 2013 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Test cases for the mock key manager.
"""

import array

from cinder import context
from cinder import exception
from cinder.keymgr import key as keymgr_key
from cinder.tests.keymgr import mock_key_mgr
from cinder.tests.keymgr import test_key_mgr


class MockKeyManagerTestCase(test_key_mgr.KeyManagerTestCase):

    def _create_key_manager(self):
        return mock_key_mgr.MockKeyManager()

    def setUp(self):
        super(MockKeyManagerTestCase, self).setUp()

        self.ctxt = context.RequestContext('fake', 'fake')

    def test_create_key(self):
        key_id_1 = self.key_mgr.create_key(self.ctxt)
        key_id_2 = self.key_mgr.create_key(self.ctxt)
        # ensure that the UUIDs are unique
        self.assertNotEqual(key_id_1, key_id_2)

    def test_create_key_with_length(self):
        for length in [64, 128, 256]:
            key_id = self.key_mgr.create_key(self.ctxt, key_length=length)
            key = self.key_mgr.get_key(self.ctxt, key_id)
            self.assertEqual(length / 8, len(key.get_encoded()))

    def test_create_null_context(self):
        self.assertRaises(exception.NotAuthorized,
                          self.key_mgr.create_key, None)

    def test_store_key(self):
        secret_key = array.array('B', ('0' * 64).decode('hex')).tolist()
        _key = keymgr_key.SymmetricKey('AES', secret_key)
        key_id = self.key_mgr.store_key(self.ctxt, _key)

        actual_key = self.key_mgr.get_key(self.ctxt, key_id)
        self.assertEqual(_key, actual_key)

    def test_store_null_context(self):
        self.assertRaises(exception.NotAuthorized,
                          self.key_mgr.store_key, None, None)

    def test_copy_key(self):
        key_id = self.key_mgr.create_key(self.ctxt)
        key = self.key_mgr.get_key(self.ctxt, key_id)

        copied_key_id = self.key_mgr.copy_key(self.ctxt, key_id)
        copied_key = self.key_mgr.get_key(self.ctxt, copied_key_id)

        self.assertNotEqual(key_id, copied_key_id)
        self.assertEqual(key, copied_key)

    def test_copy_null_context(self):
        self.assertRaises(exception.NotAuthorized,
                          self.key_mgr.copy_key, None, None)

    def test_get_key(self):
        pass

    def test_get_null_context(self):
        self.assertRaises(exception.NotAuthorized,
                          self.key_mgr.get_key, None, None)

    def test_get_unknown_key(self):
        self.assertRaises(KeyError, self.key_mgr.get_key, self.ctxt, None)

    def test_delete_key(self):
        key_id = self.key_mgr.create_key(self.ctxt)
        self.key_mgr.delete_key(self.ctxt, key_id)

        self.assertRaises(KeyError, self.key_mgr.get_key, self.ctxt, key_id)

    def test_delete_null_context(self):
        self.assertRaises(exception.NotAuthorized,
                          self.key_mgr.delete_key, None, None)

    def test_delete_unknown_key(self):
        self.assertRaises(KeyError, self.key_mgr.delete_key, self.ctxt, None)



# Copyright (c) 2011 Zadara Storage Inc.
# Copyright (c) 2011 OpenStack Foundation
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Unit Tests for volume types code
"""


import datetime
import time

from cinder import context
from cinder import db
from cinder.db.sqlalchemy import api as db_api
from cinder.db.sqlalchemy import models
from cinder import exception
from cinder.openstack.common import log as logging
from cinder import test
from cinder.tests import conf_fixture
from cinder.volume import qos_specs
from cinder.volume import volume_types


LOG = logging.getLogger(__name__)


class VolumeTypeTestCase(test.TestCase):
    """Test cases for volume type code."""
    def setUp(self):
        super(VolumeTypeTestCase, self).setUp()

        self.ctxt = context.get_admin_context()
        self.vol_type1_name = str(int(time.time()))
        self.vol_type1_specs = dict(type="physical drive",
                                    drive_type="SAS",
                                    size="300",
                                    rpm="7200",
                                    visible="True")

    def test_volume_type_create_then_destroy(self):
        """Ensure volume types can be created and deleted."""
        prev_all_vtypes = volume_types.get_all_types(self.ctxt)

        type_ref = volume_types.create(self.ctxt,
                                       self.vol_type1_name,
                                       self.vol_type1_specs)
        new = volume_types.get_volume_type_by_name(self.ctxt,
                                                   self.vol_type1_name)

        LOG.info(_("Given data: %s"), self.vol_type1_specs)
        LOG.info(_("Result data: %s"), new)

        for k, v in self.vol_type1_specs.iteritems():
            self.assertEqual(v, new['extra_specs'][k],
                             'one of fields does not match')

        new_all_vtypes = volume_types.get_all_types(self.ctxt)
        self.assertEqual(len(prev_all_vtypes) + 1,
                         len(new_all_vtypes),
                         'drive type was not created')

        volume_types.destroy(self.ctxt, type_ref['id'])
        new_all_vtypes = volume_types.get_all_types(self.ctxt)
        self.assertEqual(prev_all_vtypes,
                         new_all_vtypes,
                         'drive type was not deleted')

    def test_get_all_volume_types(self):
        """Ensures that all volume types can be retrieved."""
        session = db_api.get_session()
        total_volume_types = session.query(models.VolumeTypes).count()
        vol_types = volume_types.get_all_types(self.ctxt)
        self.assertEqual(total_volume_types, len(vol_types))

    def test_get_default_volume_type(self):
        """Ensures default volume type can be retrieved."""
        type_ref = volume_types.create(self.ctxt,
                                       conf_fixture.def_vol_type,
                                       {})
        default_vol_type = volume_types.get_default_volume_type()
        self.assertEqual(default_vol_type.get('name'),
                         conf_fixture.def_vol_type)

    def test_default_volume_type_missing_in_db(self):
        """Ensures proper exception raised if default volume type
        is not in database.
        """
        session = db_api.get_session()
        default_vol_type = volume_types.get_default_volume_type()
        self.assertEqual(default_vol_type, {})

    def test_non_existent_vol_type_shouldnt_delete(self):
        """Ensures that volume type creation fails with invalid args."""
        self.assertRaises(exception.VolumeTypeNotFound,
                          volume_types.destroy, self.ctxt, "sfsfsdfdfs")

    def test_volume_type_with_volumes_shouldnt_delete(self):
        """Ensures volume type deletion with associated volumes fail."""
        type_ref = volume_types.create(self.ctxt, self.vol_type1_name)
        db.volume_create(self.ctxt,
                         {'id': '1',
                          'updated_at': datetime.datetime(1, 1, 1, 1, 1, 1),
                          'display_description': 'Test Desc',
                          'size': 20,
                          'status': 'available',
                          'volume_type_id': type_ref['id']})
        self.assertRaises(exception.VolumeTypeInUse,
                          volume_types.destroy, self.ctxt, type_ref['id'])

    def test_repeated_vol_types_shouldnt_raise(self):
        """Ensures that volume duplicates don't raise."""
        new_name = self.vol_type1_name + "dup"
        type_ref = volume_types.create(self.ctxt, new_name)
        volume_types.destroy(self.ctxt, type_ref['id'])
        type_ref = volume_types.create(self.ctxt, new_name)

    def test_invalid_volume_types_params(self):
        """Ensures that volume type creation fails with invalid args."""
        self.assertRaises(exception.InvalidVolumeType,
                          volume_types.destroy, self.ctxt, None)
        self.assertRaises(exception.InvalidVolumeType,
                          volume_types.get_volume_type, self.ctxt, None)
        self.assertRaises(exception.InvalidVolumeType,
                          volume_types.get_volume_type_by_name,
                          self.ctxt, None)

    def test_volume_type_get_by_id_and_name(self):
        """Ensure volume types get returns same entry."""
        volume_types.create(self.ctxt,
                            self.vol_type1_name,
                            self.vol_type1_specs)
        new = volume_types.get_volume_type_by_name(self.ctxt,
                                                   self.vol_type1_name)

        new2 = volume_types.get_volume_type(self.ctxt, new['id'])
        self.assertEqual(new, new2)

    def test_volume_type_search_by_extra_spec(self):
        """Ensure volume types get by extra spec returns correct type."""
        volume_types.create(self.ctxt, "type1", {"key1": "val1",
                                                 "key2": "val2"})
        volume_types.create(self.ctxt, "type2", {"key2": "val2",
                                                 "key3": "val3"})
        volume_types.create(self.ctxt, "type3", {"key3": "another_value",
                                                 "key4": "val4"})

        vol_types = volume_types.get_all_types(
            self.ctxt,
            search_opts={'extra_specs': {"key1": "val1"}})
        LOG.info("vol_types: %s" % vol_types)
        self.assertEqual(len(vol_types), 1)
        self.assertIn("type1", vol_types.keys())
        self.assertEqual(vol_types['type1']['extra_specs'],
                         {"key1": "val1", "key2": "val2"})

        vol_types = volume_types.get_all_types(
            self.ctxt,
            search_opts={'extra_specs': {"key2": "val2"}})
        LOG.info("vol_types: %s" % vol_types)
        self.assertEqual(len(vol_types), 2)
        self.assertIn("type1", vol_types.keys())
        self.assertIn("type2", vol_types.keys())

        vol_types = volume_types.get_all_types(
            self.ctxt,
            search_opts={'extra_specs': {"key3": "val3"}})
        LOG.info("vol_types: %s" % vol_types)
        self.assertEqual(len(vol_types), 1)
        self.assertIn("type2", vol_types.keys())

    def test_volume_type_search_by_extra_spec_multiple(self):
        """Ensure volume types get by extra spec returns correct type."""
        volume_types.create(self.ctxt, "type1", {"key1": "val1",
                                                 "key2": "val2",
                                                 "key3": "val3"})
        volume_types.create(self.ctxt, "type2", {"key2": "val2",
                                                 "key3": "val3"})
        volume_types.create(self.ctxt, "type3", {"key1": "val1",
                                                 "key3": "val3",
                                                 "key4": "val4"})

        vol_types = volume_types.get_all_types(
            self.ctxt,
            search_opts={'extra_specs': {"key1": "val1",
                                         "key3": "val3"}})
        LOG.info("vol_types: %s" % vol_types)
        self.assertEqual(len(vol_types), 2)
        self.assertIn("type1", vol_types.keys())
        self.assertIn("type3", vol_types.keys())
        self.assertEqual(vol_types['type1']['extra_specs'],
                         {"key1": "val1", "key2": "val2", "key3": "val3"})
        self.assertEqual(vol_types['type3']['extra_specs'],
                         {"key1": "val1", "key3": "val3", "key4": "val4"})

    def test_is_encrypted(self):
        volume_type = volume_types.create(self.ctxt, "type1")
        volume_type_id = volume_type.get('id')
        self.assertFalse(volume_types.is_encrypted(self.ctxt, volume_type_id))

        encryption = {
            'control_location': 'front-end',
            'provider': 'fake_provider',
        }
        db_api.volume_type_encryption_create(self.ctxt, volume_type_id,
                                             encryption)
        self.assertTrue(volume_types.is_encrypted(self.ctxt, volume_type_id))

    def test_get_volume_type_qos_specs(self):
        qos_ref = qos_specs.create(self.ctxt, 'qos-specs-1', {'k1': 'v1',
                                                              'k2': 'v2',
                                                              'k3': 'v3'})
        type_ref = volume_types.create(self.ctxt, "type1", {"key2": "val2",
                                                  "key3": "val3"})
        res = volume_types.get_volume_type_qos_specs(type_ref['id'])
        self.assertIsNone(res['qos_specs'])
        qos_specs.associate_qos_with_type(self.ctxt,
                                          qos_ref['id'],
                                          type_ref['id'])

        expected = {'qos_specs': {'id': qos_ref['id'],
                                  'name': 'qos-specs-1',
                                  'consumer': 'back-end',
                                  'specs': {
                                      'k1': 'v1',
                                      'k2': 'v2',
                                      'k3': 'v3'}}}
        res = volume_types.get_volume_type_qos_specs(type_ref['id'])
        self.assertDictMatch(expected, res)

    def test_volume_types_diff(self):
        #type_ref 1 and 2 have the same extra_specs, while 3 has different
        keyvals1 = {"key1": "val1", "key2": "val2"}
        keyvals2 = {"key1": "val0", "key2": "val2"}
        type_ref1 = volume_types.create(self.ctxt, "type1", keyvals1)
        type_ref2 = volume_types.create(self.ctxt, "type2", keyvals1)
        type_ref3 = volume_types.create(self.ctxt, "type3", keyvals2)

        # Check equality with only extra_specs
        diff, same = volume_types.volume_types_diff(self.ctxt, type_ref1['id'],
                                                    type_ref2['id'])
        self.assertEqual(same, True)
        self.assertEqual(diff['extra_specs']['key1'], ('val1', 'val1'))
        diff, same = volume_types.volume_types_diff(self.ctxt, type_ref1['id'],
                                                    type_ref3['id'])
        self.assertEqual(same, False)
        self.assertEqual(diff['extra_specs']['key1'], ('val1', 'val0'))

        #qos_ref 1 and 2 have the same specs, while 3 has different
        qos_keyvals1 = {'k1': 'v1', 'k2': 'v2', 'k3': 'v3'}
        qos_keyvals2 = {'k1': 'v0', 'k2': 'v2', 'k3': 'v3'}
        qos_ref1 = qos_specs.create(self.ctxt, 'qos-specs-1', qos_keyvals1)
        qos_ref2 = qos_specs.create(self.ctxt, 'qos-specs-2', qos_keyvals1)
        qos_ref3 = qos_specs.create(self.ctxt, 'qos-specs-3', qos_keyvals2)

        # Check equality with qos specs too
        qos_specs.associate_qos_with_type(self.ctxt, qos_ref1['id'],
                                          type_ref1['id'])
        qos_specs.associate_qos_with_type(self.ctxt, qos_ref2['id'],
                                          type_ref2['id'])
        diff, same = volume_types.volume_types_diff(self.ctxt, type_ref1['id'],
                                                    type_ref2['id'])
        self.assertEqual(same, True)
        self.assertEqual(diff['extra_specs']['key1'], ('val1', 'val1'))
        self.assertEqual(diff['qos_specs']['k1'], ('v1', 'v1'))
        qos_specs.disassociate_qos_specs(self.ctxt, qos_ref2['id'],
                                         type_ref2['id'])
        qos_specs.associate_qos_with_type(self.ctxt, qos_ref3['id'],
                                          type_ref2['id'])
        diff, same = volume_types.volume_types_diff(self.ctxt, type_ref1['id'],
                                                    type_ref2['id'])
        self.assertEqual(same, False)
        self.assertEqual(diff['extra_specs']['key1'], ('val1', 'val1'))
        self.assertEqual(diff['qos_specs']['k1'], ('v1', 'v0'))
        qos_specs.disassociate_qos_specs(self.ctxt, qos_ref3['id'],
                                         type_ref2['id'])
        qos_specs.associate_qos_with_type(self.ctxt, qos_ref2['id'],
                                          type_ref2['id'])

        # And add encryption for good measure
        enc_keyvals1 = {'cipher': 'c1', 'key_size': 256, 'provider': 'p1',
                        'control_location': 'front-end'}
        enc_keyvals2 = {'cipher': 'c1', 'key_size': 128, 'provider': 'p1',
                        'control_location': 'front-end'}
        db.volume_type_encryption_create(self.ctxt, type_ref1['id'],
                                         enc_keyvals1)
        db.volume_type_encryption_create(self.ctxt, type_ref2['id'],
                                         enc_keyvals2)
        diff, same = volume_types.volume_types_diff(self.ctxt, type_ref1['id'],
                                                    type_ref2['id'])
        self.assertEqual(same, False)
        self.assertEqual(diff['extra_specs']['key1'], ('val1', 'val1'))
        self.assertEqual(diff['qos_specs']['k1'], ('v1', 'v1'))
        self.assertEqual(diff['encryption']['key_size'], (256, 128))

        # Check diff equals type specs when one type is None
        diff, same = volume_types.volume_types_diff(self.ctxt, None,
                                                    type_ref1['id'])
        self.assertEqual(same, False)
        self.assertEqual(diff['extra_specs'],
                         {'key1': (None, 'val1'), 'key2': (None, 'val2')})
        self.assertEqual(diff['qos_specs'],
                         {'consumer': (None, 'back-end'),
                          'k1': (None, 'v1'),
                          'k2': (None, 'v2'),
                          'k3': (None, 'v3')})
        self.assertEqual(diff['encryption'],
                         {'cipher': (None, 'c1'),
                          'control_location': (None, 'front-end'),
                          'deleted': (None, False),
                          'key_size': (None, 256),
                          'provider': (None, 'p1')})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import traceback

from oslo.config import cfg
import taskflow.engines
from taskflow.patterns import linear_flow
from taskflow.utils import misc

from cinder import exception
from cinder import flow_utils
from cinder.image import glance
from cinder.openstack.common import log as logging
from cinder.openstack.common import processutils
from cinder.openstack.common import timeutils
from cinder import utils
from cinder.volume.flows import common
from cinder.volume import utils as volume_utils

LOG = logging.getLogger(__name__)

ACTION = 'volume:create'
CONF = cfg.CONF

# These attributes we will attempt to save for the volume if they exist
# in the source image metadata.
IMAGE_ATTRIBUTES = (
    'checksum',
    'container_format',
    'disk_format',
    'min_disk',
    'min_ram',
    'size',
)


class OnFailureRescheduleTask(flow_utils.CinderTask):
    """Triggers a rescheduling request to be sent when reverting occurs.

    Reversion strategy: Triggers the rescheduling mechanism whereby a cast gets
    sent to the scheduler rpc api to allow for an attempt X of Y for scheduling
    this volume elsewhere.
    """

    def __init__(self, reschedule_context, db, scheduler_rpcapi):
        requires = ['filter_properties', 'image_id', 'request_spec',
                    'snapshot_id', 'volume_id', 'context']
        super(OnFailureRescheduleTask, self).__init__(addons=[ACTION],
                                                      requires=requires)
        self.scheduler_rpcapi = scheduler_rpcapi
        self.db = db
        self.reschedule_context = reschedule_context
        # These exception types will trigger the volume to be set into error
        # status rather than being rescheduled.
        self.no_reschedule_types = [
            # Image copying happens after volume creation so rescheduling due
            # to copy failure will mean the same volume will be created at
            # another place when it still exists locally.
            exception.ImageCopyFailure,
            # Metadata updates happen after the volume has been created so if
            # they fail, rescheduling will likely attempt to create the volume
            # on another machine when it still exists locally.
            exception.MetadataCopyFailure,
            exception.MetadataCreateFailure,
            exception.MetadataUpdateFailure,
            # The volume/snapshot has been removed from the database, that
            # can not be fixed by rescheduling.
            exception.VolumeNotFound,
            exception.SnapshotNotFound,
            exception.VolumeTypeNotFound,
            exception.ImageUnacceptable,
        ]

    def execute(self, **kwargs):
        pass

    def _reschedule(self, context, cause, request_spec, filter_properties,
                    snapshot_id, image_id, volume_id, **kwargs):
        """Actions that happen during the rescheduling attempt occur here."""

        create_volume = self.scheduler_rpcapi.create_volume
        if not filter_properties:
            filter_properties = {}
        if 'retry' not in filter_properties:
            filter_properties['retry'] = {}

        retry_info = filter_properties['retry']
        num_attempts = retry_info.get('num_attempts', 0)
        request_spec['volume_id'] = volume_id

        LOG.debug(_("Volume %(volume_id)s: re-scheduling %(method)s "
                    "attempt %(num)d due to %(reason)s") %
                  {'volume_id': volume_id,
                   'method': common.make_pretty_name(create_volume),
                   'num': num_attempts,
                   'reason': cause.exception_str})

        if all(cause.exc_info):
            # Stringify to avoid circular ref problem in json serialization
            retry_info['exc'] = traceback.format_exception(*cause.exc_info)

        return create_volume(context, CONF.volume_topic, volume_id,
                             snapshot_id=snapshot_id, image_id=image_id,
                             request_spec=request_spec,
                             filter_properties=filter_properties)

    def _post_reschedule(self, context, volume_id):
        """Actions that happen after the rescheduling attempt occur here."""

        LOG.debug(_("Volume %s: re-scheduled"), volume_id)

    def _pre_reschedule(self, context, volume_id):
        """Actions that happen before the rescheduling attempt occur here."""

        try:
            # Reset the volume state.
            #
            # NOTE(harlowja): this is awkward to be done here, shouldn't
            # this happen at the scheduler itself and not before it gets
            # sent to the scheduler? (since what happens if it never gets
            # there??). It's almost like we need a status of 'on-the-way-to
            # scheduler' in the future.
            update = {
                'status': 'creating',
                'scheduled_at': timeutils.utcnow(),
            }
            LOG.debug(_("Updating volume %(volume_id)s with %(update)s.") %
                      {'update': update, 'volume_id': volume_id})
            self.db.volume_update(context, volume_id, update)
        except exception.CinderException:
            # Don't let resetting the status cause the rescheduling to fail.
            LOG.exception(_("Volume %s: resetting 'creating' status failed."),
                          volume_id)

    def revert(self, context, result, flow_failures, **kwargs):
        # Check if we have a cause which can tell us not to reschedule.
        for failure in flow_failures.values():
            if failure.check(*self.no_reschedule_types):
                return

        volume_id = kwargs['volume_id']
        # Use a different context when rescheduling.
        if self.reschedule_context:
            context = self.reschedule_context
            try:
                cause = list(flow_failures.values())[0]
                self._pre_reschedule(context, volume_id)
                self._reschedule(context, cause, **kwargs)
                self._post_reschedule(context, volume_id)
            except exception.CinderException:
                LOG.exception(_("Volume %s: rescheduling failed"), volume_id)


class ExtractVolumeRefTask(flow_utils.CinderTask):
    """Extracts volume reference for given volume id."""

    default_provides = 'volume_ref'

    def __init__(self, db, host):
        super(ExtractVolumeRefTask, self).__init__(addons=[ACTION])
        self.db = db
        self.host = host

    def execute(self, context, volume_id):
        # NOTE(harlowja): this will fetch the volume from the database, if
        # the volume has been deleted before we got here then this should fail.
        #
        # In the future we might want to have a lock on the volume_id so that
        # the volume can not be deleted while its still being created?
        volume_ref = self.db.volume_get(context, volume_id)

        # NOTE(vish): so we don't have to get volume from db again before
        # passing it to the driver.
        volume_ref['host'] = self.host

        return volume_ref

    def revert(self, context, volume_id, result, **kwargs):
        if isinstance(result, misc.Failure):
            return

        common.error_out_volume(context, self.db, volume_id)
        LOG.error(_("Volume %s: create failed"), volume_id)


class ExtractVolumeSpecTask(flow_utils.CinderTask):
    """Extracts a spec of a volume to be created into a common structure.

    This task extracts and organizes the input requirements into a common
    and easier to analyze structure for later tasks to use. It will also
    attach the underlying database volume reference which can be used by
    other tasks to reference for further details about the volume to be.

    Reversion strategy: N/A
    """

    default_provides = 'volume_spec'

    def __init__(self, db):
        requires = ['image_id', 'snapshot_id', 'source_volid']
        super(ExtractVolumeSpecTask, self).__init__(addons=[ACTION],
                                                    requires=requires)
        self.db = db

    def execute(self, context, volume_ref, **kwargs):
        get_remote_image_service = glance.get_remote_image_service

        volume_name = volume_ref['name']
        volume_size = utils.as_int(volume_ref['size'], quiet=False)

        # Create a dictionary that will represent the volume to be so that
        # later tasks can easily switch between the different types and create
        # the volume according to the volume types specifications (which are
        # represented in this dictionary).
        specs = {
            'status': volume_ref['status'],
            'type': 'raw',  # This will have the type of the volume to be
                            # created, which should be one of [raw, snap,
                            # source_vol, image]
            'volume_id': volume_ref['id'],
            'volume_name': volume_name,
            'volume_size': volume_size,
        }

        if kwargs.get('snapshot_id'):
            # We are making a snapshot based volume instead of a raw volume.
            specs.update({
                'type': 'snap',
                'snapshot_id': kwargs['snapshot_id'],
            })
        elif kwargs.get('source_volid'):
            # We are making a source based volume instead of a raw volume.
            #
            # NOTE(harlowja): This will likely fail if the source volume
            # disappeared by the time this call occurred.
            source_volid = kwargs['source_volid']
            source_volume_ref = self.db.volume_get(context, source_volid)
            specs.update({
                'source_volid': source_volid,
                # This is captured incase we have to revert and we want to set
                # back the source volume status to its original status. This
                # may or may not be sketchy to do??
                'source_volstatus': source_volume_ref['status'],
                'type': 'source_vol',
            })
        elif kwargs.get('image_id'):
            # We are making a image based volume instead of a raw volume.
            image_href = kwargs['image_id']
            image_service, image_id = get_remote_image_service(context,
                                                               image_href)
            specs.update({
                'type': 'image',
                'image_id': image_id,
                'image_location': image_service.get_location(context,
                                                             image_id),
                'image_meta': image_service.show(context, image_id),
                # Instead of refetching the image service later just save it.
                #
                # NOTE(harlowja): if we have to later recover this tasks output
                # on another 'node' that this object won't be able to be
                # serialized, so we will have to recreate this object on
                # demand in the future.
                'image_service': image_service,
            })

        return specs

    def revert(self, context, result, **kwargs):
        if isinstance(result, misc.Failure):
            return
        volume_spec = result.get('volume_spec')
        # Restore the source volume status and set the volume to error status.
        common.restore_source_status(context, self.db, volume_spec)


class NotifyVolumeActionTask(flow_utils.CinderTask):
    """Performs a notification about the given volume when called.

    Reversion strategy: N/A
    """

    def __init__(self, db, event_suffix):
        super(NotifyVolumeActionTask, self).__init__(addons=[ACTION,
                                                             event_suffix])
        self.db = db
        self.event_suffix = event_suffix

    def execute(self, context, volume_ref):
        volume_id = volume_ref['id']
        try:
            volume_utils.notify_about_volume_usage(context, volume_ref,
                                                   self.event_suffix,
                                                   host=volume_ref['host'])
        except exception.CinderException:
            # If notification sending of volume database entry reading fails
            # then we shouldn't error out the whole workflow since this is
            # not always information that must be sent for volumes to operate
            LOG.exception(_("Failed notifying about the volume"
                            " action %(event)s for volume %(volume_id)s") %
                          {'event': self.event_suffix,
                           'volume_id': volume_id})


class CreateVolumeFromSpecTask(flow_utils.CinderTask):
    """Creates a volume from a provided specification.

    Reversion strategy: N/A
    """

    default_provides = 'volume'

    def __init__(self, db, driver):
        super(CreateVolumeFromSpecTask, self).__init__(addons=[ACTION])
        self.db = db
        self.driver = driver
        # This maps the different volume specification types into the methods
        # that can create said volume type (aka this is a jump table).
        self._create_func_mapping = {
            'raw': self._create_raw_volume,
            'snap': self._create_from_snapshot,
            'source_vol': self._create_from_source_volume,
            'image': self._create_from_image,
        }

    def _handle_bootable_volume_glance_meta(self, context, volume_id,
                                            **kwargs):
        """Enable bootable flag and properly handle glance metadata.

        Caller should provide one and only one of snapshot_id,source_volid
        and image_id. If an image_id specified, a image_meta should also be
        provided, otherwise will be treated as an empty dictionary.
        """

        log_template = _("Copying metadata from %(src_type)s %(src_id)s to "
                         "%(vol_id)s.")
        exception_template = _("Failed updating volume %(vol_id)s metadata"
                               " using the provided %(src_type)s"
                               " %(src_id)s metadata")
        src_type = None
        src_id = None
        self._enable_bootable_flag(context, volume_id)
        try:
            if kwargs.get('snapshot_id'):
                src_type = 'snapshot'
                src_id = kwargs['snapshot_id']
                snapshot_id = src_id
                LOG.debug(log_template % {'src_type': src_type,
                                          'src_id': src_id,
                                          'vol_id': volume_id})
                self.db.volume_glance_metadata_copy_to_volume(
                    context, volume_id, snapshot_id)
            elif kwargs.get('source_volid'):
                src_type = 'source volume'
                src_id = kwargs['source_volid']
                source_volid = src_id
                LOG.debug(log_template % {'src_type': src_type,
                                          'src_id': src_id,
                                          'vol_id': volume_id})
                self.db.volume_glance_metadata_copy_from_volume_to_volume(
                    context,
                    source_volid,
                    volume_id)
            elif kwargs.get('image_id'):
                src_type = 'image'
                src_id = kwargs['image_id']
                image_id = src_id
                image_meta = kwargs.get('image_meta', {})
                LOG.debug(log_template % {'src_type': src_type,
                                          'src_id': src_id,
                                          'vol_id': volume_id})
                self._capture_volume_image_metadata(context, volume_id,
                                                    image_id, image_meta)
        except exception.CinderException as ex:
            LOG.exception(exception_template % {'src_type': src_type,
                                                'src_id': src_id,
                                                'vol_id': volume_id})
            raise exception.MetadataCopyFailure(reason=ex)

    def _create_from_snapshot(self, context, volume_ref, snapshot_id,
                              **kwargs):
        volume_id = volume_ref['id']
        snapshot_ref = self.db.snapshot_get(context, snapshot_id)
        model_update = self.driver.create_volume_from_snapshot(volume_ref,
                                                               snapshot_ref)
        # NOTE(harlowja): Subtasks would be useful here since after this
        # point the volume has already been created and further failures
        # will not destroy the volume (although they could in the future).
        make_bootable = False
        try:
            originating_vref = self.db.volume_get(context,
                                                  snapshot_ref['volume_id'])
            make_bootable = originating_vref.bootable
        except exception.CinderException as ex:
            LOG.exception(_("Failed fetching snapshot %(snapshot_id)s bootable"
                            " flag using the provided glance snapshot "
                            "%(snapshot_ref_id)s volume reference") %
                          {'snapshot_id': snapshot_id,
                           'snapshot_ref_id': snapshot_ref['volume_id']})
            raise exception.MetadataUpdateFailure(reason=ex)
        if make_bootable:
            self._handle_bootable_volume_glance_meta(context, volume_id,
                                                     snapshot_id=snapshot_id)
        return model_update

    def _enable_bootable_flag(self, context, volume_id):
        try:
            LOG.debug(_('Marking volume %s as bootable.'), volume_id)
            self.db.volume_update(context, volume_id, {'bootable': True})
        except exception.CinderException as ex:
            LOG.exception(_("Failed updating volume %(volume_id)s bootable"
                            " flag to true") % {'volume_id': volume_id})
            raise exception.MetadataUpdateFailure(reason=ex)

    def _create_from_source_volume(self, context, volume_ref,
                                   source_volid, **kwargs):
        # NOTE(harlowja): if the source volume has disappeared this will be our
        # detection of that since this database call should fail.
        #
        # NOTE(harlowja): likely this is not the best place for this to happen
        # and we should have proper locks on the source volume while actions
        # that use the source volume are underway.
        srcvol_ref = self.db.volume_get(context, source_volid)
        model_update = self.driver.create_cloned_volume(volume_ref, srcvol_ref)
        # NOTE(harlowja): Subtasks would be useful here since after this
        # point the volume has already been created and further failures
        # will not destroy the volume (although they could in the future).
        if srcvol_ref.bootable:
            self._handle_bootable_volume_glance_meta(context, volume_ref['id'],
                                                     source_volid=source_volid)
        return model_update

    def _copy_image_to_volume(self, context, volume_ref,
                              image_id, image_location, image_service):
        """Downloads Glance image to the specified volume."""
        copy_image_to_volume = self.driver.copy_image_to_volume
        volume_id = volume_ref['id']
        LOG.debug(_("Attempting download of %(image_id)s (%(image_location)s)"
                    " to volume %(volume_id)s.") %
                  {'image_id': image_id, 'volume_id': volume_id,
                   'image_location': image_location})
        try:
            copy_image_to_volume(context, volume_ref, image_service, image_id)
        except processutils.ProcessExecutionError as ex:
            LOG.error(_("Failed to copy image %(image_id)s to volume: "
                        "%(volume_id)s, error: %(error)s") %
                      {'volume_id': volume_id,
                       'error': ex.stderr, 'image_id': image_id})
            raise exception.ImageCopyFailure(reason=ex.stderr)
        except exception.ImageUnacceptable as ex:
            LOG.error(_("Failed to copy image to volume: %(volume_id)s, "
                        "error: %(error)s") % {'volume_id': volume_id,
                                               'error': ex})
            raise exception.ImageUnacceptable(ex)
        except Exception as ex:
            LOG.error(_("Failed to copy image %(image_id)s to "
                        "volume: %(volume_id)s, error: %(error)s") %
                      {'volume_id': volume_id, 'error': ex,
                       'image_id': image_id})
            if not isinstance(ex, exception.ImageCopyFailure):
                raise exception.ImageCopyFailure(reason=ex)
            else:
                raise

        LOG.debug(_("Downloaded image %(image_id)s (%(image_location)s)"
                    " to volume %(volume_id)s successfully.") %
                  {'image_id': image_id, 'volume_id': volume_id,
                   'image_location': image_location})

    def _capture_volume_image_metadata(self, context, volume_id,
                                       image_id, image_meta):

        # Save some base attributes into the volume metadata
        base_metadata = {
            'image_id': image_id,
        }
        name = image_meta.get('name', None)
        if name:
            base_metadata['image_name'] = name

        # Save some more attributes into the volume metadata from the image
        # metadata
        for key in IMAGE_ATTRIBUTES:
            if key not in image_meta:
                continue
            value = image_meta.get(key, None)
            if value is not None:
                base_metadata[key] = value

        # Save all the image metadata properties into the volume metadata
        property_metadata = {}
        image_properties = image_meta.get('properties', {})
        for (key, value) in image_properties.items():
            if value is not None:
                property_metadata[key] = value

        # NOTE(harlowja): The best way for this to happen would be in bulk,
        # but that doesn't seem to exist (yet), so we go through one by one
        # which means we can have partial create/update failure.
        volume_metadata = dict(property_metadata)
        volume_metadata.update(base_metadata)
        LOG.debug(_("Creating volume glance metadata for volume %(volume_id)s"
                    " backed by image %(image_id)s with: %(vol_metadata)s.") %
                  {'volume_id': volume_id, 'image_id': image_id,
                   'vol_metadata': volume_metadata})
        for (key, value) in volume_metadata.items():
            try:
                self.db.volume_glance_metadata_create(context, volume_id,
                                                      key, value)
            except exception.GlanceMetadataExists:
                pass

    def _create_from_image(self, context, volume_ref,
                           image_location, image_id, image_meta,
                           image_service, **kwargs):
        LOG.debug(_("Cloning %(volume_id)s from image %(image_id)s "
                    " at location %(image_location)s.") %
                  {'volume_id': volume_ref['id'],
                   'image_location': image_location, 'image_id': image_id})
        # Create the volume from an image.
        #
        # NOTE (singn): two params need to be returned
        # dict containing provider_location for cloned volume
        # and clone status.
        model_update, cloned = self.driver.clone_image(
            volume_ref, image_location, image_id, image_meta)
        if not cloned:
            # TODO(harlowja): what needs to be rolled back in the clone if this
            # volume create fails?? Likely this should be a subflow or broken
            # out task in the future. That will bring up the question of how
            # do we make said subflow/task which is only triggered in the
            # clone image 'path' resumable and revertable in the correct
            # manner.
            #
            # Create the volume and then download the image onto the volume.
            model_update = self.driver.create_volume(volume_ref)
            updates = dict(model_update or dict(), status='downloading')
            try:
                volume_ref = self.db.volume_update(context,
                                                   volume_ref['id'], updates)
            except exception.CinderException:
                LOG.exception(_("Failed updating volume %(volume_id)s with "
                                "%(updates)s") %
                              {'volume_id': volume_ref['id'],
                               'updates': updates})
            self._copy_image_to_volume(context, volume_ref,
                                       image_id, image_location, image_service)

        self._handle_bootable_volume_glance_meta(context, volume_ref['id'],
                                                 image_id=image_id,
                                                 image_meta=image_meta)
        return model_update

    def _create_raw_volume(self, context, volume_ref, **kwargs):
        return self.driver.create_volume(volume_ref)

    def execute(self, context, volume_ref, volume_spec):
        volume_spec = dict(volume_spec)
        volume_id = volume_spec.pop('volume_id', None)

        # we can't do anything if the driver didn't init
        if not self.driver.initialized:
            driver_name = self.driver.__class__.__name__
            LOG.error(_("Unable to create volume. "
                        "Volume driver %s not initialized") % driver_name)
            # NOTE(flaper87): Set the error status before
            # raising any exception.
            self.db.volume_update(context, volume_id, dict(status='error'))
            raise exception.DriverNotInitialized()

        create_type = volume_spec.pop('type', None)
        create_functor = self._create_func_mapping.get(create_type)
        if not create_functor:
            raise exception.VolumeTypeNotFound(volume_type_id=create_type)

        if not volume_id:
            volume_id = volume_ref['id']
        LOG.info(_("Volume %(volume_id)s: being created using %(functor)s "
                   "with specification: %(volume_spec)s") %
                 {'volume_spec': volume_spec, 'volume_id': volume_id,
                  'functor': common.make_pretty_name(create_functor)})

        # Call the given functor to make the volume.
        model_update = create_functor(context, volume_ref=volume_ref,
                                      **volume_spec)

        # Persist any model information provided on creation.
        try:
            if model_update:
                volume_ref = self.db.volume_update(context, volume_ref['id'],
                                                   model_update)
        except exception.CinderException:
            # If somehow the update failed we want to ensure that the
            # failure is logged (but not try rescheduling since the volume at
            # this point has been created).
            LOG.exception(_("Failed updating model of volume %(volume_id)s"
                            " with creation provided model %(model)s") %
                          {'volume_id': volume_id, 'model': model_update})
            raise

        return volume_ref


class CreateVolumeOnFinishTask(NotifyVolumeActionTask):
    """On successful volume creation this will perform final volume actions.

    When a volume is created successfully it is expected that MQ notifications
    and database updates will occur to 'signal' to others that the volume is
    now ready for usage. This task does those notifications and updates in a
    reliable manner (not re-raising exceptions if said actions can not be
    triggered).

    Reversion strategy: N/A
    """

    def __init__(self, db, event_suffix):
        super(CreateVolumeOnFinishTask, self).__init__(db, event_suffix)
        self.status_translation = {
            'migration_target_creating': 'migration_target',
        }

    def execute(self, context, volume, volume_spec):
        volume_id = volume['id']
        new_status = self.status_translation.get(volume_spec.get('status'),
                                                 'available')
        update = {
            'status': new_status,
            'launched_at': timeutils.utcnow(),
        }
        try:
            # TODO(harlowja): is it acceptable to only log if this fails??
            # or are there other side-effects that this will cause if the
            # status isn't updated correctly (aka it will likely be stuck in
            # 'building' if this fails)??
            volume_ref = self.db.volume_update(context, volume_id, update)
            # Now use the parent to notify.
            super(CreateVolumeOnFinishTask, self).execute(context, volume_ref)
        except exception.CinderException:
            LOG.exception(_("Failed updating volume %(volume_id)s with "
                            "%(update)s") % {'volume_id': volume_id,
                                             'update': update})
        # Even if the update fails, the volume is ready.
        msg = _("Volume %(volume_name)s (%(volume_id)s): created successfully")
        LOG.info(msg % {
            'volume_name': volume_spec['volume_name'],
            'volume_id': volume_id,
        })


def get_flow(context, db, driver, scheduler_rpcapi, host, volume_id,
             allow_reschedule, reschedule_context, request_spec,
             filter_properties, snapshot_id=None, image_id=None,
             source_volid=None):
    """Constructs and returns the manager entrypoint flow.

    This flow will do the following:

    1. Determines if rescheduling is enabled (ahead of time).
    2. Inject keys & values for dependent tasks.
    3. Selects 1 of 2 activated only on *failure* tasks (one to update the db
       status & notify or one to update the db status & notify & *reschedule*).
    4. Extracts a volume specification from the provided inputs.
    5. Notifies that the volume has start to be created.
    6. Creates a volume from the extracted volume specification.
    7. Attaches a on-success *only* task that notifies that the volume creation
       has ended and performs further database status updates.
    """

    flow_name = ACTION.replace(":", "_") + "_manager"
    volume_flow = linear_flow.Flow(flow_name)

    # This injects the initial starting flow values into the workflow so that
    # the dependency order of the tasks provides/requires can be correctly
    # determined.
    create_what = {
        'context': context,
        'filter_properties': filter_properties,
        'image_id': image_id,
        'request_spec': request_spec,
        'snapshot_id': snapshot_id,
        'source_volid': source_volid,
        'volume_id': volume_id,
    }

    volume_flow.add(ExtractVolumeRefTask(db, host))

    if allow_reschedule and request_spec:
        volume_flow.add(OnFailureRescheduleTask(reschedule_context,
                                                db, scheduler_rpcapi))

    volume_flow.add(ExtractVolumeSpecTask(db),
                    NotifyVolumeActionTask(db, "create.start"),
                    CreateVolumeFromSpecTask(db, driver),
                    CreateVolumeOnFinishTask(db, "create.end"))

    # Now load (but do not run) the flow using the provided initial data.
    return taskflow.engines.load(volume_flow, store=create_what)
